@@ -18,25 +18,60 @@ from cluster_vcf_records import allele_combinations, utils, vcf_file_read, vcf_r
 Variant = namedtuple("Variant", ["seq_id", "pos", "ref", "alt"])
 
 
+def _run_one_cluster_chunk(
+    tracker, outprefix, max_ref_length, max_alleles, index_range
+):
+    logging.info(f"Start clustering chunk {index_range[0]}-{index_range[1]}")
+    tracker.cluster_one_chunk(
+        outprefix,
+        max_ref_length,
+        max_alleles=max_alleles,
+        split_start=index_range[0],
+        split_end=index_range[1],
+    )
+
+
+def vcf_records_make_same_allele_combination(record1, record2, ref_seqs):
+    if record1.CHROM != record2.CHROM:
+        return False
+    assert record1.ref_end_pos() < record2.POS
+    alleles1 = record1.inferred_var_seqs_plus_flanks(ref_seqs[record1.CHROM], 0)[1]
+    alleles2 = record2.inferred_var_seqs_plus_flanks(ref_seqs[record2.CHROM], 0)[1]
+    join_seq = ref_seqs[record1.CHROM][record1.ref_end_pos() + 1 : record2.POS]
+
+    combinations = set()
+    for a1 in alleles1:
+        for a2 in alleles2:
+            allele = a1 + join_seq + a2
+            if allele in combinations:
+                return True
+            combinations.add(allele)
+    return False
+
+
 def variants_overlap(var1, var2):
     end1 = var1.pos + len(var1.ref) - 1
     end2 = var2.pos + len(var2.ref) - 1
     return var1.seq_id == var2.seq_id and var1.pos <= end2 and var2.pos <= end1
 
 
-def _load_one_vcf_file(vcf_file, ref_seqs, ref_seq_to_id, ref_fasta, temp_dir):
+def _load_one_vcf_file(
+    vcf_file, ref_seqs, ref_seq_to_id, ref_fasta, temp_dir, break_alleles
+):
     sample = vcf_file_read.get_sample_name_from_vcf_file(vcf_file)
     if sample is None:
         raise Exception(f"Error getting sample name from vcf file {vcf_file}")
     tmpdir = tempfile.mkdtemp(prefix="normalize_vcf.", dir=temp_dir)
     simplified_vcf = os.path.join(tmpdir, "simplified.vcf")
     normalized_vcf = os.path.join(tmpdir, "normalized.vcf")
-    utils.simplify_vcf(vcf_file, simplified_vcf)
-    utils.normalise_vcf(simplified_vcf, ref_fasta, normalized_vcf)
+    utils.simplify_vcf(vcf_file, simplified_vcf, ref_seqs=ref_seqs)
+    utils.normalise_vcf(
+        simplified_vcf, ref_fasta, normalized_vcf, break_alleles=break_alleles
+    )
     os.unlink(simplified_vcf)
     variants = []
 
-    with open(normalized_vcf) as f:
+    with vcf_file_read.open_vcf_file_for_reading(normalized_vcf) as f:
         for line in f:
             if line.startswith("#"):
                 continue
@@ -265,6 +300,8 @@ class VariantTracker:
         self.ref_fasta = os.path.abspath(ref_fasta)
         self.mem_lmit = mem_limit
         self.variants = Variants()
+        self.cluster_limit = 8
+        self.max_var_gap = 10
         (
             self.ref_seqs,
             self.ref_seq_names,
@@ -294,6 +331,7 @@ class VariantTracker:
     def _add_var_block_file(self):
         filename = f"block.{len(self.var_block_files)}.tsv.gz"
         self.var_block_files.append(filename)
+        self.var_block_tabixes.append(os.path.join(self.root_dir, filename))
         self.samples.append([])
 
     def _write_last_var_block_file(self, var_block):
@@ -322,6 +360,7 @@ class VariantTracker:
         mem_limit=2,
         force=False,
         sample_limit=None,
+        break_alleles=True,
     ):
         if force:
             utils.rm_rf(self.root_dir)
@@ -336,6 +375,7 @@ class VariantTracker:
             made_temp_dir = True
 
         self.var_block_files = []
+        self.var_block_tabixes = []
         self.samples = []
         self._add_var_block_file()
         self.variants = Variants()
@@ -343,9 +383,7 @@ class VariantTracker:
 
         for i in range(0, len(infiles), cpus):
             with multiprocessing.Pool(cpus) as pool:
-                new_variants_lists = (
-                    self.var_block_filesew_variants_lists
-                ) = pool.starmap(
+                new_variants_lists = pool.starmap(
                     _load_one_vcf_file,
                     zip(
                         infiles[i : i + cpus],
@@ -353,6 +391,7 @@ class VariantTracker:
                         itertools.repeat(self.ref_seq_to_id),
                         itertools.repeat(self.ref_fasta),
                         itertools.repeat(temp_dir),
+                        itertools.repeat(break_alleles),
                     ),
                 )
 
@@ -395,6 +434,8 @@ class VariantTracker:
         logging.info("Finished merging VCF files")
 
     def _variant_cluster_to_vcf_line(self, variants, variant_ids, max_alleles=None):
+        if len(variants) == 0:
+            return None
         ref_seq = self.ref_seqs[self.ref_seq_names[variants[0].seq_id]]
         if logging.getLogger().level <= logging.DEBUG:
             logging.debug(f"Clustering variants:")
@@ -432,50 +473,262 @@ class VariantTracker:
             info_field = "High_variability"
 
         if len(alts) == 0:
+            logging.warning("Could not make VCF record from these variants:")
+            for variant in variants:
+                logging.warning("    " + str(variant))
             return None
         else:
-            return "\t".join(
-                [
-                    ref_seq.id,
-                    str(start + 1),
-                    ".",
-                    ref_seq[start : end + 1],
-                    ",".join(sorted(list(alts))),
-                    ".",
-                    "PASS",
-                    info_field,
-                ]
+            return vcf_record.VcfRecord(
+                "\t".join(
+                    [
+                        ref_seq.id,
+                        str(start + 1),
+                        ".",
+                        ref_seq[start : end + 1],
+                        ",".join(sorted(list(alts))),
+                        ".",
+                        "PASS",
+                        info_field,
+                    ]
+                )
             )
 
-    def cluster(self, outprefix, max_ref_length, max_alleles=None):
+    def _print_cluster_vcf_header(self, filehandle, max_alleles):
+        print("##fileformat=VCFv4.2", file=filehandle)
+        for seq in self.ref_seqs.values():
+            print(f"##contig=<ID={seq.id},length={len(seq)}>", file=filehandle)
+        print('##FILTER=<ID=PASS,Description="All filters passed">', file=filehandle)
+        if max_alleles is not None:
+            print(
+                f'##INFO=<ID=High_variability,Number=0,Type=Flag,Description="Position is in a region of high variability, with more alleles than the limit of {max_alleles}">',
+                file=filehandle,
+            )
+        print("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT", file=filehandle)
+
+    def _init_clusters(self):
+        self.clusters = []
+        self.clusters_insertions = set()
+        self.clusters_deletions = set()
+
+    def _append_to_clusters(self, to_append):
+        has_insertion = any([len(to_append[0].REF) < len(x) for x in to_append[0].ALT])
+        has_deletion = any([len(to_append[0].REF) > len(x) for x in to_append[0].ALT])
+        if has_insertion:
+            self.clusters_insertions.add(len(self.clusters))
+        if has_deletion:
+            self.clusters_deletions.add(len(self.clusters))
+        self.clusters.append(to_append)
+
+    def _merge_last_records_in_clusters(self, max_alleles):
+        if (
+            len(self.clusters) < 2
+            or len(self.clusters_insertions) == 0
+            or len(self.clusters_deletions) == 0
+        ):
+            return
+
+        best_merged = None
+        best_merged_index = None
+        merged_record, merged_vars, merged_var_ids = self.clusters[-1]
+        for i in reversed(range(len(self.clusters) - 1)):
+            if (
+                merged_record.CHROM != self.clusters[i][0].CHROM
+                or self.clusters[i][0].ref_end_pos() + self.max_var_gap
+                < merged_record.POS
+            ):
+                break
+
+            have_same_combos = vcf_records_make_same_allele_combination(
+                self.clusters[i][0], merged_record, self.ref_seqs
+            )
+            if i == 0 and not have_same_combos:
+                break
+
+            merged_vars = self.clusters[i][1] + merged_vars
+            merged_var_ids = self.clusters[i][2] + merged_var_ids
+            merged_record = self._variant_cluster_to_vcf_line(
+                merged_vars, merged_var_ids, max_alleles=max_alleles
+            )
+            if merged_record is None:
+                break
+
+            if have_same_combos:
+                best_merged = (merged_record, merged_vars, merged_var_ids)
+                best_merged_index = i
+
+        if best_merged_index is not None:
+            self.clusters_insertions = {
+                x for x in self.clusters_insertions if x < best_merged_index
+            }
+            self.clusters_deletions = {
+                x for x in self.clusters_deletions if x < best_merged_index
+            }
+            del self.clusters[best_merged_index:]
+            self._append_to_clusters(best_merged)
+
+    def _merge_last_records_in_clusters_old(self, max_alleles):
+        while len(self.clusters) > 1:
+            if vcf_records_make_same_allele_combination(
+                self.clusters[-2][0], self.clusters[-1][0], self.ref_seqs
+            ):
+                new_variants = self.clusters[-2][1] + self.clusters[-1][1]
+                new_variant_ids = self.clusters[-2][2] + self.clusters[-1][2]
+                new_record = self._variant_cluster_to_vcf_line(
+                    new_variants, new_variant_ids, max_alleles=max_alleles
+                )
+                if new_record is not None:
+                    self.clusters.pop()
+                    self.clusters[-1] = (new_record, new_variants, new_variant_ids)
+                else:
+                    break
+            else:
+                break
+
+    def _print_first_clusters(self, filehandle):
+        while len(self.clusters) > self.cluster_limit:
+            print(self.clusters[0][0], file=filehandle)
+            self.clusters.pop(0)
+            self.clusters_insertions.discard(0)
+            self.clusters_deletions.discard(0)
+            self.clusters_insertions = {x - 1 for x in self.clusters_insertions}
+            self.clusters_deletions = {x - 1 for x in self.clusters_deletions}
+
+    def find_parallel_cluster_split_points(self, max_ref_length, splits):
+        targets = [x * int(len(self.variants) / splits) for x in range(1, splits)]
+        good_count = 0
+        previous_var = None
+        split_ranges = []
+        looking_for_split = False
+        target_index = 0
+        finishing_split = None
+
+        mask = {}  # ref id -> set of bad genome coords because in indels
+        for i, (var, var_id) in enumerate(self.variants.sorted_iter()):
+            if len(var.ref) > max_ref_length:
+                continue
+
+            if len(var.ref) > 1 or len(var.alt) > 1:
+                if var.seq_id not in mask:
+                    mask[var.seq_id] = set()
+                for pos in range(var.pos, var.pos + len(var.ref)):
+                    mask[var.seq_id].add(pos)
+
+        for i, (var, var_id) in enumerate(self.variants.sorted_iter()):
+            if len(var.ref) > max_ref_length:
+                continue
+
+            var_in_mask = False
+            if var.seq_id in mask:
+                for pos in range(var.pos, var.pos + len(var.ref)):
+                    if pos in mask[var.seq_id]:
+                        var_in_mask = True
+                        break
+
+            if finishing_split is not None:
+                if var.pos == finishing_split:
+                    assert not var_in_mask
+                    split_ranges[-1][-1] = i
+                else:
+                    finishing_split = None
+                    if target_index >= len(targets):
+                        break
+
+            if not looking_for_split and i > targets[target_index]:
+                if len(var.ref) == len(var.alt) == 1 and not var_in_mask:
+                    looking_for_split = True
+                    good_count = 1
+                    previous_var = var
+                    continue
+
+            if looking_for_split:
+                if len(var.ref) == len(var.alt) == 1 and not var_in_mask:
+                    if previous_var is None or var.pos > previous_var.pos:
+                        good_count += 1
+                    previous_var = var
+                else:
+                    good_count = 0
+                    previous_var = None
+
+                if good_count > self.cluster_limit + 3:
+                    if len(split_ranges) == 0:
+                        split_ranges.append([0, i])
+                    else:
+                        split_ranges.append([split_ranges[-1][-1] + 1, i])
+                    finishing_split = var.pos
+                    target_index += 1
+                    looking_for_split = False
+
+        split_ranges.append([split_ranges[-1][-1] + 1, len(self.variants) - 1])
+        return split_ranges
+
+    def cluster(self, outprefix, max_ref_length, max_alleles=None, cpus=1):
+        if cpus > 1:
+            split_ranges = self.find_parallel_cluster_split_points(max_ref_length, cpus)
+            outprefixes = [f"{outprefix}.{i}" for i in range(len(split_ranges))]
+            logging.info(f"Running clustering in parallel using {cpus} cpus")
+            with multiprocessing.Pool(cpus) as pool:
+                pool.starmap(
+                    _run_one_cluster_chunk,
+                    zip(
+                        itertools.repeat(self),
+                        outprefixes,
+                        itertools.repeat(max_ref_length),
+                        itertools.repeat(max_alleles),
+                        split_ranges,
+                    ),
+                )
+            utils.cat_vcfs(
+                [x + ".vcf" for x in outprefixes], f"{outprefix}.vcf", delete_files=True
+            )
+            utils.cat_tsvs(
+                [x + ".excluded.tsv" for x in outprefixes],
+                f"{outprefix}.excluded.tsv",
+                delete_files=True,
+            )
+        else:
+            self.cluster_one_chunk(outprefix, max_ref_length, max_alleles=max_alleles)
+
+    def cluster_one_chunk(
+        self,
+        outprefix,
+        max_ref_length,
+        max_alleles=None,
+        split_start=None,
+        split_end=None,
+    ):
         out_main = f"{outprefix}.vcf"
         out_exclude = f"{outprefix}.excluded.tsv"
+        total_vars = (
+            len(self.variants) if split_start is None else split_end - split_start + 1
+        )
 
         with open(out_main, "w") as f_main, open(out_exclude, "w") as f_exclude:
-            print("##fileformat=VCFv4.2", file=f_main)
-            for seq in self.ref_seqs.values():
-                print(f"##contig=<ID={seq.id},length={len(seq)}>", file=f_main)
-            print('##FILTER=<ID=PASS,Description="All filters passed">', file=f_main)
-            if max_alleles is not None:
-                print(
-                    f'##INFO=<ID=High_variability,Number=0,Type=Flag,Description="Position is in a region of high variability, with more alleles than the limit of {max_alleles}">',
-                    file=f_main,
-                )
-            print("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO", file=f_main)
+            self._print_cluster_vcf_header(f_main, max_alleles)
             print("CHROM\tPOS\tREF\tALT", file=f_exclude)
             variants = []
             variant_ids = []
             var_count = 0
-            report_count = int(len(self.variants) / 20)
+            report_count = int(total_vars / 10)
+            self._init_clusters()
+
+            for i, (var, var_id) in enumerate(self.variants.sorted_iter()):
+                if split_start is not None:
+                    if i < split_start:
+                        continue
+                    elif i > split_end:
+                        break
 
-            for var, var_id in self.variants.sorted_iter():
                 var_count += 1
                 if var_count >= report_count:
-                    pc_complete = round(100 * var_count / len(self.variants))
+                    pc_complete = round(100 * var_count / total_vars)
+                    if split_start is None:
+                        range_string = ""
+                    else:
+                        range_string = f" for split range {split_start}-{split_end}"
                     logging.info(
-                        f"Completed {pc_complete}% ({var_count}/{len(self.variants)}) of variants"
+                        f"Completed {pc_complete}% ({var_count}/{total_vars}) of variants{range_string}"
                     )
-                    report_count += int(len(self.variants) / 20)
+                    report_count += int(total_vars / 10)
 
                 if len(var.ref) > max_ref_length:
                     print(
@@ -498,8 +751,16 @@ class VariantTracker:
                 record = self._variant_cluster_to_vcf_line(
                     variants, variant_ids, max_alleles=max_alleles
                 )
-                if record is not None:
-                    print(record, file=f_main)
+                if record is None:
+                    variants = []
+                    variants_ids = []
+                    continue
+                else:
+                    self._append_to_clusters((record, variants, variant_ids))
+
+                self._merge_last_records_in_clusters(max_alleles)
+                self._print_first_clusters(f_main)
+
                 variants = [var]
                 variant_ids = [var_id]
 
@@ -507,4 +768,7 @@ class VariantTracker:
                 variants, variant_ids, max_alleles=max_alleles
             )
             if record is not None:
-                print(record, file=f_main)
+                self._append_to_clusters((record, variants, variant_ids))
+            self._merge_last_records_in_clusters(max_alleles)
+            for c in self.clusters:
+                print(c[0], file=f_main)
